from dataclasses import dataclass
from logging import Logger, getLogger
from typing import Any, Generic, List, Optional, Sequence, Union

from cachetools import Cache
from circuitbreaker import CircuitBreakerError

from dbdaora.exceptions import EntityNotFoundError

from ..circuitbreaker import AsyncCircuitBreaker
from ..entity import Entity, EntityData
from ..keys import FallbackKey
from ..repository import MemoryRepository


@dataclass(init=False)
class Service(Generic[Entity, EntityData, FallbackKey]):
    repository: MemoryRepository[Entity, EntityData, FallbackKey]
    circuit_breaker: AsyncCircuitBreaker
    cache: Optional[Cache]
    logger: Logger

    def __init__(
        self,
        repository: MemoryRepository[Entity, EntityData, FallbackKey],
        circuit_breaker: AsyncCircuitBreaker,
        cache: Optional[Cache] = None,
        exists_cache: Optional[Cache] = None,
        logger: Logger = getLogger(__name__),
    ):
        self.repository = repository
        self.circuit_breaker = circuit_breaker
        self.cache = cache
        self.exists_cache = exists_cache
        self.entity_circuit = self.circuit_breaker(self.repository.entity)
        self.entities_circuit = self.circuit_breaker(self.repository.entities)
        self.add_circuit = self.circuit_breaker(self.repository.add)
        self.delete_circuit = self.circuit_breaker(self.repository.delete)
        self.exists_circuit = self.circuit_breaker(self.repository.exists)
        self.logger = logger

    async def get_many(
        self,
        *ids: str,
        fields: Optional[Sequence[str]] = None,
        **filters: Any,
    ) -> Sequence[Any]:
        try:
            if self.cache is None:
                return [
                    entity
                    for entity in await self.entities_circuit(
                        self.repository.query(
                            many=ids, fields=fields, **filters
                        )
                    )
                    if entity is not None
                ]

            return await self.get_many_cached(
                ids, self.cache, fields=fields, **filters
            )

        except CircuitBreakerError as error:
            self.logger.warning(error)
            if self.cache is not None:
                return await self.get_many_cached(
                    ids, self.cache, fields=fields, memory=False, **filters
                )

            return [
                entity
                for entity in await self.repository.query(
                    many=ids, fields=fields, memory=False, **filters
                ).entities
                if entity is not None
            ]

    async def get_many_cached(
        self,
        ids: Sequence[str],
        cache: Cache,
        fields: Optional[Sequence[str]] = None,
        memory: bool = True,
        **filters: Any,
    ) -> Sequence[Any]:
        missed_ids: List[str] = []
        entities = {}
        cache_key_suffix = self.cache_key_suffix(**filters)

        for id_ in ids:
            entity = self.get_cached_entity(id_, cache_key_suffix, fields)

            if entity is None:
                missed_ids.append(id_)

            entities[id_] = entity

        if missed_ids:
            try:
                if memory:
                    missed_entities = await self.entities_circuit(
                        self.repository.query(
                            many=missed_ids, fields=fields, **filters
                        )
                    )
                else:
                    missed_entities = await self.repository.query(
                        many=missed_ids, fields=fields, memory=False, **filters
                    ).entities
            except EntityNotFoundError:
                missed_entities = []

            for entity in missed_entities:
                id_ = (
                    entity[self.repository.id_name]
                    if isinstance(entity, dict)
                    else entity
                    if isinstance(entity, str)
                    else getattr(entity, self.repository.id_name)
                )
                entities[id_] = entity

        final_entities = []

        for id_, entity in entities.items():
            if entity is not None and entity is not CACHE_ALREADY_NOT_FOUND:
                final_entities.append(entity)
                self.set_cached_entity(id_, cache_key_suffix, entity)

            elif entity is None:
                self.set_cached_entity(
                    id_, cache_key_suffix, CACHE_ALREADY_NOT_FOUND
                )

        if not final_entities:
            raise EntityNotFoundError(ids, fields, filters)

        return final_entities

    def get_cached_entity(
        self, id: str, key_suffix: str, fields: Optional[Sequence[str]] = None,
    ) -> Any:
        if self.cache is None:
            return None

        entity = self.cache.get(self.cache_key(id, key_suffix))

        if fields is None:
            return entity

        if isinstance(entity, dict):
            for field in fields:
                if field not in entity:
                    return None

        else:
            for field in fields:
                if hasattr(entity, field):
                    return None

        return entity

    def cache_key(self, id: str, suffix: str) -> str:
        return f'{id}{suffix}'

    def set_cached_entity(
        self,
        id: str,
        key_suffix: str,
        entity: Union[Entity, 'CacheAlreadyNotFound'],
    ) -> None:
        if self.cache is not None:
            self.cache[self.cache_key(id, key_suffix)] = entity

    def cache_key_suffix(self, **filters: Any) -> str:
        return (
            ''.join(f'{f}{v}' for f, v in filters.items()) if filters else ''
        )

    async def get_one(
        self, id: str, fields: Optional[Sequence[str]] = None, **filters: Any
    ) -> Any:
        try:
            if self.cache is None:
                return await self.entity_circuit(
                    self.repository.query(id=id, fields=fields, **filters)
                )

            return await self.get_one_cached(
                id, self.cache, fields=fields, **filters
            )

        except CircuitBreakerError as error:
            self.logger.warning(error)
            if self.cache is not None:
                return await self.get_one_cached(
                    id, self.cache, fields=fields, memory=False, **filters
                )

            return await self.repository.query(
                id=id, fields=fields, memory=False, **filters
            ).entity

    async def get_one_cached(
        self,
        id: str,
        cache: Cache,
        fields: Optional[Sequence[str]] = None,
        memory: bool = True,
        **filters: Any,
    ) -> Any:
        cache_key_suffix = self.cache_key_suffix(**filters)
        entity = self.get_cached_entity(id, cache_key_suffix, fields)

        if entity is None:
            try:
                if memory:
                    entity = await self.entity_circuit(
                        self.repository.query(id=id, fields=fields, **filters)
                    )
                else:
                    entity = await self.repository.query(
                        id=id, fields=fields, memory=False, **filters
                    ).entity

                self.set_cached_entity(id, cache_key_suffix, entity)

            except EntityNotFoundError:
                self.set_cached_entity(
                    id, cache_key_suffix, CACHE_ALREADY_NOT_FOUND
                )
                raise

        elif entity == CACHE_ALREADY_NOT_FOUND:
            raise EntityNotFoundError(id)

        return entity

    async def add(
        self, entity: Any, *entities: Any, memory: bool = True
    ) -> None:
        if not memory:
            await self.repository.add(entity, *entities, memory=False)
            return

        try:
            await self.add_circuit(entity, *entities)

        except CircuitBreakerError as error:
            self.logger.warning(error)
            await self.repository.add(entity, *entities, memory=False)

    async def delete(self, entity_id: str, **filters: Any) -> None:
        try:
            await self.delete_circuit(
                self.repository.query(id=entity_id, **filters)
            )

        except CircuitBreakerError as error:
            self.logger.warning(error)
            await self.repository.query(
                id=entity_id, memory=False, **filters
            ).delete

    async def exists(self, id: str, **filters: Any) -> bool:
        try:
            if self.exists_cache is None:
                return await self.exists_circuit(
                    self.repository.query(id=id, **filters)
                )

            return await self.exists_cached(id, self.exists_cache, **filters)

        except CircuitBreakerError as error:
            self.logger.warning(error)
            if self.exists_cache is not None:
                return await self.exists_cached(
                    id, self.exists_cache, memory=False, **filters,
                )

            return await self.repository.query(
                id=id, memory=False, **filters
            ).exists

    async def exists_cached(
        self, id: str, cache: Cache, memory: bool = True, **filters: Any,
    ) -> bool:
        cache_key = self.cache_key(id, self.cache_key_suffix(**filters))
        entity_exists = cache.get(cache_key)

        if entity_exists is None:
            if memory:
                entity_exists = await self.exists_circuit(
                    self.repository.query(id=id, **filters)
                )
            else:
                entity_exists = await self.repository.query(
                    id=id, memory=False, **filters
                ).exists

            if not entity_exists:
                cache[cache_key] = CACHE_ALREADY_NOT_FOUND
                return False
            else:
                cache[cache_key] = True

        elif entity_exists == CACHE_ALREADY_NOT_FOUND:
            return False

        return True

    async def shutdown(self) -> None:
        self.repository.memory_data_source.close()
        await self.repository.memory_data_source.wait_closed()


class CacheAlreadyNotFound:
    ...


CACHE_ALREADY_NOT_FOUND = CacheAlreadyNotFound()



from typing import Any, ClassVar, Dict, Optional, Sequence, Tuple, Union

from .. import DataSource


RangeOutput = Union[
    Sequence[Tuple[str, float]],
    Sequence[Tuple[bytes, float]],
    Sequence[Union[bytes, str]],
]


class MemoryMultiExec:
    def delete(self, key: str) -> Any:
        raise NotImplementedError()  # pragma: no cover

    def hmset(
        self,
        key: str,
        field: Union[str, bytes],
        value: Union[str, bytes],
        *pairs: Union[str, bytes],
    ) -> Any:
        raise NotImplementedError()  # pragma: no cover

    def zadd(
        self, key: str, score: float, member: str, *pairs: Union[float, str]
    ) -> Any:
        raise NotImplementedError()  # pragma: no cover

    async def execute(self, *, return_exceptions: bool = False) -> Any:
        raise NotImplementedError()  # pragma: no cover


class MemoryDataSource(DataSource):
    key_separator: ClassVar[str] = ':'

    def make_key(self, *key_parts: str) -> str:
        return self.key_separator.join(key_parts)

    async def get(self, key: str) -> Optional[bytes]:
        raise NotImplementedError()  # pragma: no cover

    async def set(self, key: str, data: str) -> None:
        raise NotImplementedError()  # pragma: no cover

    async def delete(self, key: str) -> None:
        raise NotImplementedError()  # pragma: no cover

    async def expire(self, key: str, time: int) -> None:
        raise NotImplementedError()  # pragma: no cover

    async def exists(self, key: str) -> int:
        raise NotImplementedError()  # pragma: no cover

    async def zrevrange(
        self, key: str, withscores: bool = False
    ) -> Optional[RangeOutput]:
        raise NotImplementedError()  # pragma: no cover

    async def zrange(
        self, key: str, withscores: bool = False
    ) -> Optional[RangeOutput]:
        raise NotImplementedError()  # pragma: no cover

    async def zadd(
        self, key: str, score: float, member: str, *pairs: Union[float, str]
    ) -> None:
        raise NotImplementedError()  # pragma: no cover

    async def hmset(
        self,
        key: str,
        field: Union[str, bytes],
        value: Union[str, bytes],
        *pairs: Union[str, bytes],
    ) -> None:
        raise NotImplementedError()  # pragma: no cover

    async def hmget(
        self, key: str, field: Union[str, bytes], *fields: Union[str, bytes]
    ) -> Sequence[Optional[bytes]]:
        raise NotImplementedError()  # pragma: no cover

    async def hgetall(self, key: str) -> Dict[bytes, bytes]:
        raise NotImplementedError()  # pragma: no cover

    def close(self) -> None:
        raise NotImplementedError()  # pragma: no cover

    async def wait_closed(self) -> None:
        raise NotImplementedError()  # pragma: no cover

    def multi_exec(self) -> MemoryMultiExec:
        raise NotImplementedError()  # pragma: no cover



import asyncio
import dataclasses
from typing import Any, Dict, List, Optional, Sequence, Union

from . import MemoryDataSource, MemoryMultiExec, RangeOutput


@dataclasses.dataclass
class DictMultiExec(MemoryMultiExec):
    client: 'DictMemoryDataSource'
    futures: List[Any] = dataclasses.field(default_factory=list)

    def delete(self, key: str) -> Any:
        future = self.client.delete(key)
        self.futures.append(future)
        return future

    def hmset(
        self,
        key: str,
        field: Union[str, bytes],
        value: Union[str, bytes],
        *pairs: Union[str, bytes],
    ) -> Any:
        future = self.client.hmset(key, field, value, *pairs)
        self.futures.append(future)
        return future

    def zadd(
        self, key: str, score: float, member: str, *pairs: Union[float, str]
    ) -> Any:
        future = self.client.zadd(key, score, member, *pairs)
        self.futures.append(future)
        return future

    async def execute(self, *, return_exceptions: bool = False) -> Any:
        results = await asyncio.gather(*self.futures)
        self.futures.clear()
        return results


@dataclasses.dataclass
class DictMemoryDataSource(MemoryDataSource):
    db: Dict[str, Any] = dataclasses.field(default_factory=dict)

    async def get(self, key: str) -> Optional[bytes]:
        return self.db.get(key)

    async def set(self, key: str, data: str) -> None:
        self.db[key] = data.encode()

    async def delete(self, key: str) -> None:
        self.db.pop(key, None)

    async def expire(self, key: str, time: int) -> None:
        ...

    async def exists(self, key: str) -> bool:
        return key in self.db

    async def zrevrange(
        self, key: str, withscores: bool = False
    ) -> Optional[RangeOutput]:
        data: RangeOutput

        if key not in self.db:
            return None

        if withscores:
            data = list(self.db[key])
        else:
            data = [i[0] for i in self.db[key]]

        data.reverse()
        return data

    async def zrange(
        self, key: str, withscores: bool = False
    ) -> Optional[RangeOutput]:
        data: Optional[RangeOutput] = self.db.get(key)

        if data is None:
            return None

        if withscores:
            return data

        return [i[0] for i in self.db[key]]

    async def zadd(
        self, key: str, score: float, member: str, *pairs: Union[float, str]
    ) -> None:
        data = [score, member] + list(pairs)
        self.db[key] = sorted(
            [(data[i], data[i - 1]) for i in range(1, len(data), 2)],
            key=lambda d: d[1],
        )

    async def hmset(
        self,
        key: str,
        field: Union[str, bytes],
        value: Union[str, bytes],
        *pairs: Union[str, bytes],
    ) -> None:
        data = [field, value] + list(pairs)
        self.db[key] = {
            f.encode()
            if isinstance(f := data[i - 1], str)  # noqa
            else (f if isinstance(f, bytes) else str(f).encode()): v.encode()
            if isinstance(v := data[i], str)  # noqa
            else (v if isinstance(v, bytes) else str(v).encode())
            for i in range(1, len(data), 2)
        }

    async def hmget(
        self, key: str, field: Union[str, bytes], *fields: Union[str, bytes]
    ) -> Sequence[Optional[bytes]]:
        data: Optional[Dict[bytes, Any]] = self.db.get(key)

        if data is None:
            return [None for i in range(len(fields) + 1)]

        return [
            None
            if (d := data.get(f.encode() if isinstance(f, str) else f))  # noqa
            is None
            else (
                d
                if isinstance(d, bytes)
                else (d.encode() if isinstance(d, str) else str(d).encode())
            )
            for f in (field,) + fields
        ]

    async def hgetall(self, key: str) -> Dict[bytes, bytes]:
        return {
            f: d.encode()
            if isinstance(d, str)
            else (
                d
                if isinstance(d, bytes)
                else (d.encode() if isinstance(d, str) else str(d).encode())
            )
            for f, d in self.db.get(key, {}).items()
        }

    def multi_exec(self) -> MemoryMultiExec:
        return DictMultiExec(self)



"""Communicates with databases using repository pattern and service patterns"""


__version__ = '0.15.1'

from dbdaora.cache import CacheType, TTLDaoraCache
from dbdaora.circuitbreaker import AsyncCircuitBreaker
from dbdaora.data_sources.fallback import FallbackDataSource
from dbdaora.data_sources.fallback.dict import DictFallbackDataSource
from dbdaora.data_sources.memory import MemoryDataSource
from dbdaora.data_sources.memory.dict import DictMemoryDataSource
from dbdaora.exceptions import EntityNotFoundError
from dbdaora.hash.factory import make_service as make_hash_service
from dbdaora.hash.query import HashQuery, HashQueryMany
from dbdaora.hash.repositories import HashData, HashRepository
from dbdaora.hash.service import HashService
from dbdaora.hashring import HashRing
from dbdaora.query import Query, QueryMany
from dbdaora.repository import MemoryRepository
from dbdaora.service import Service
from dbdaora.service.builder import build as build_service
from dbdaora.service.builder import build_cache
from dbdaora.sorted_set.entity import (
    SortedSetData,
    SortedSetDictEntity,
    SortedSetEntity,
)
from dbdaora.sorted_set.query import SortedSetQuery
from dbdaora.sorted_set.repositories import SortedSetRepository


from dbdaora.boolean.repositories import BooleanRepository  # noqa isort:skip
from dbdaora.boolean.service import BooleanService  # noqa isort:skip
from dbdaora.boolean.repositories.datastore import (  # noqa isort:skip
    DatastoreBooleanRepository,
)

try:
    from dbdaora.data_sources.fallback.datastore import DatastoreDataSource
    from dbdaora.hash.repositories.datastore import DatastoreHashRepository
    from dbdaora.sorted_set.repositories.datastore import (
        DatastoreSortedSetRepository,
    )
except ImportError:
    DatastoreDataSource = None  # type: ignore
    DatastoreHashRepository = None  # type: ignore
    DatastoreSortedSetRepository = None  # type: ignore

try:
    from dbdaora.data_sources.memory.aioredis import (
        AioRedisDataSource,
        ShardsAioRedisDataSource,
        make as make_aioredis_data_source,
    )
except ImportError:
    AioRedisDataSource = None  # type: ignore
    ShardsAioRedisDataSource = None  # type: ignore
    make_aioredis_data_source = None  # type: ignore

try:
    from dbdaora.data_sources.fallback.mongodb import (
        MongoDataSource,
        Key as MongoKey,
    )
    from dbdaora.hash.repositories.mongodb import MongodbHashRepository
    from dbdaora.hash.service import MongoHashService
except ImportError:
    MongoDataSource = None  # type: ignore
    MongodbHashRepository = None  # type: ignore


__all__ = [
    'MemoryRepository',
    'HashRepository',
    'HashQuery',
    'HashData',
    'SortedSetRepository',
    'SortedSetQuery',
    'SortedSetEntity',
    'DictFallbackDataSource',
    'HashService',
    'AsyncCircuitBreaker',
    'HashRing',
    'FallbackDataSource',
    'MemoryDataSource',
    'DictMemoryDataSource',
    'build_service',
    'CacheType',
    'Service',
    'EntityNotFoundError',
    'HashQueryMany',
    'make_hash_service',
    'SortedSetData',
    'SortedSetDictEntity',
    'TTLDaoraCache',
    'build_cache',
    'BooleanRepository',
    'DatastoreBooleanRepository',
    'BooleanService',
    'Query',
    'QueryMany',
]

if AioRedisDataSource:
    __all__.append('AioRedisDataSource')

if ShardsAioRedisDataSource:
    __all__.append('ShardsAioRedisDataSource')

if make_aioredis_data_source:
    __all__.append('make_aioredis_data_source')

if DatastoreDataSource:
    __all__.append('DatastoreDataSource')

if DatastoreHashRepository:
    __all__.append('DatastoreHashRepository')

if DatastoreSortedSetRepository:
    __all__.append('DatastoreSortedSetRepository')

if MongoDataSource:
    __all__.append('MongoDataSource')

if MongodbHashRepository:
    __all__.append('MongodbHashRepository')

if MongoHashService:
    __all__.append('MongoHashService')

if MongoKey:
    __all__.append('MongoKey')



from google.cloud.datastore import Key

from ..data_sources.fallback.mongodb import Key as MongoKey
from ..entity import Entity
from ..keys import FallbackKey
from ..service import Service
from .repositories import HashData


class HashService(Service[Entity, HashData, FallbackKey]):
    ...


class DatastoreHashService(Service[Entity, HashData, Key]):
    ...


class MongoHashService(Service[Entity, HashData, MongoKey]):
    ...



from datetime import datetime
from typing import Any, Awaitable, Callable, Optional, Type, TypeVar

from circuitbreaker import (
    STATE_CLOSED,
    STATE_OPEN,
    CircuitBreaker,
    CircuitBreakerError,
)


FuncReturn = TypeVar('FuncReturn')


class AsyncCircuitBreaker(CircuitBreaker):
    _last_failure: Optional[Exception]

    def __init__(
        self,
        failure_threshold: Optional[int] = None,
        recovery_timeout: Optional[int] = None,
        expected_exception: Optional[Type[Exception]] = None,
        name: Optional[str] = None,
        fallback_function: Optional[
            Callable[..., Awaitable[FuncReturn]]
        ] = None,
    ):
        self._last_failure = None
        self._failure_count = 0
        self._failure_threshold = (
            failure_threshold
            if failure_threshold is not None
            else self.FAILURE_THRESHOLD
        )
        self._recovery_timeout = recovery_timeout or self.RECOVERY_TIMEOUT
        self._expected_exception = (
            expected_exception or self.EXPECTED_EXCEPTION
        )
        self._fallback_function = fallback_function or self.FALLBACK_FUNCTION
        self._name = name
        self._state = STATE_CLOSED
        self._opened = datetime.utcnow()

    async def call(
        self,
        func: Callable[..., Awaitable[FuncReturn]],
        *args: Any,
        **kwargs: Any,
    ) -> FuncReturn:
        if self.opened:
            if self.fallback_function:
                await self.fallback_function(*args, **kwargs)
            else:
                raise CircuitBreakerError(self)
        try:
            result = await func(*args, **kwargs)
        except self._expected_exception as e:
            self._last_failure = e
            self.__call_failed()

            if self._failure_threshold == 0:
                raise CircuitBreakerError(self)

            raise

        self.__call_succeeded()
        return result

    def __call_succeeded(self) -> None:
        self._state = STATE_CLOSED
        self._last_failure = None
        self._failure_count = 0

    def __call_failed(self) -> None:
        self._failure_count += 1
        if self._failure_count >= self._failure_threshold:
            self._state = STATE_OPEN
            self._opened = datetime.utcnow()



import dataclasses
from typing import Any, Dict, Iterable, Optional

import motor.motor_asyncio as motor

from . import FallbackDataSource


@dataclasses.dataclass
class Key:
    collection_name: str
    document_id: str


@dataclasses.dataclass
class MongoDataSource(FallbackDataSource[Key]):
    database_name: str
    client: motor.AsyncIOMotorClient = dataclasses.field(
        default_factory=motor.AsyncIOMotorClient
    )

    def make_key(self, *key_parts: Any) -> Key:
        return Key(
            key_parts[0],
            self.key_separator.join([str(k) for k in key_parts[1:]]),
        )

    async def get(self, key: Key) -> Optional[Dict[str, Any]]:
        collection = self.collection(key)
        return await collection.find_one({'_id': key.document_id})

    async def put(
        self,
        key: Key,
        data: Dict[str, Any],
        exclude_from_indexes: Iterable[str] = (),
        **kwargs: Any,
    ) -> None:
        collection = self.collection(key)
        await collection.replace_one(
            {'_id': key.document_id}, data, upsert=True,
        )

    async def delete(self, key: Key) -> None:
        collection = self.collection(key)
        await collection.delete_one({'_id': key.document_id})

    def collection(self, key: Key) -> motor.AsyncIOMotorCollection:
        return self.client[self.database_name][key.collection_name]



import dataclasses
from typing import Any, List, Optional

from dbdaora.keys import FallbackKey
from dbdaora.query import Query

from .entity import Entity, SortedSetData


@dataclasses.dataclass
class SortedSetQuery(Query[Entity, SortedSetData, FallbackKey]):
    repository: 'SortedSetRepository[FallbackKey]'
    reverse: bool = False
    withscores: bool = False
    page: Optional[int] = None
    page_size: Optional[int] = None
    min: Optional[float] = None
    max: Optional[float] = None

    def __init__(
        self,
        repository: 'SortedSetRepository[FallbackKey]',
        *args: Any,
        memory: bool = True,
        key_parts: Optional[List[Any]] = None,
        reverse: bool = False,
        withscores: bool = False,
        page: Optional[int] = None,
        page_size: Optional[int] = None,
        min: Optional[float] = None,
        max: Optional[float] = None,
        **kwargs: Any,
    ):
        super().__init__(
            repository, memory=memory, key_parts=key_parts, *args, **kwargs,
        )
        self.reverse = reverse
        self.withscores = withscores
        self.page = page
        self.page_size = page_size
        self.min = min
        self.max = max


from .repositories import SortedSetRepository  # noqa isort:skip



from typing import Any, Dict, Protocol, Sequence, TypedDict, Union

from dbdaora.data_sources.memory import RangeOutput


SortedSetInput = Sequence[Union[str, float]]

SortedSetData = Union[RangeOutput, SortedSetInput]


class SortedSetEntity(Protocol):
    @property
    def values(self) -> SortedSetData:
        raise NotImplementedError()  # pragma: no cover

    def __init__(self, values: SortedSetData, **kwargs: Any):
        raise NotImplementedError()  # pragma: no cover


class SortedSetDictEntity(TypedDict):
    values: SortedSetData


# see https://github.com/python/mypy/issues/4300
# Entity = Union[SortedSetEntity, SortedSetDictEntity]
Entity = Union[SortedSetEntity, Dict[str, Any]]



import itertools
from typing import Any, Optional, Sequence, Tuple, TypedDict, Union

from dbdaora.keys import FallbackKey
from dbdaora.repository import MemoryRepository

from ..entity import SortedSetData
from ..query import SortedSetQuery


class FallbackSortedSetData(TypedDict):
    values: Sequence[Tuple[str, float]]


class SortedSetRepository(MemoryRepository[Any, SortedSetData, FallbackKey],):
    __skip_cls_validation__ = ('SortedSetRepository',)

    async def get_memory_data(  # type: ignore
        self, key: str, query: SortedSetQuery[FallbackKey],
    ) -> Optional[SortedSetData]:
        if query.reverse:
            return await self.memory_data_source.zrevrange(key)

        return await self.memory_data_source.zrange(key)

    async def get_fallback_data(
        self,
        query: Union[SortedSetQuery[FallbackKey], Any],
        for_memory: bool = False,
    ) -> Optional[SortedSetData]:
        data: Optional[FallbackSortedSetData]

        data = await self.fallback_data_source.get(  # type: ignore
            self.fallback_key(query)
        )

        if data is None:
            return None

        elif for_memory or (
            isinstance(query, SortedSetQuery) and query.withscores
        ):
            return [  # type: ignore
                (data['values'][i], data['values'][i + 1])
                for i in range(0, len(data['values']), 2)
            ]

        return [data['values'][i] for i in range(0, len(data['values']), 2)]

    def make_entity(  # type: ignore
        self, data: SortedSetData, query: SortedSetQuery[FallbackKey]
    ) -> Any:
        return self.get_entity_type(query)(
            values=data,
            **{self.id_name: query.attribute_from_key(self.id_name)},
        )

    def make_entity_from_fallback(  # type: ignore
        self, data: SortedSetData, query: SortedSetQuery[FallbackKey]
    ) -> Any:
        return self.make_entity(data, query)

    async def add_memory_data(self, key: str, data: SortedSetData) -> None:
        transaction = self.memory_data_source.multi_exec()
        transaction.delete(key)
        transaction.zadd(key, *data)
        await transaction.execute()

    async def add_fallback(
        self, entity: Any, *entities: Any, **kwargs: Any
    ) -> None:
        await self.fallback_data_source.put(
            self.fallback_key(entity),
            {
                'values': list(itertools.chain(*entity['values']))
                if isinstance(entity, dict)
                else list(itertools.chain(*entity.values))
            },
            **kwargs,
        )

    async def add_memory_data_from_fallback(  # type: ignore
        self,
        key: str,
        query: Union[SortedSetQuery[FallbackKey], Any],
        data: Sequence[Tuple[str, float]],
    ) -> SortedSetData:
        await self.memory_data_source.zadd(key, *self.format_memory_data(data))

        if isinstance(query, SortedSetQuery) and query.withscores:
            return data

        return [i[0] for i in data]

    def make_query(
        self, *args: Any, **kwargs: Any
    ) -> SortedSetQuery[FallbackKey]:
        return SortedSetQuery(self, *args, **kwargs)

    def make_memory_data_from_entity(self, entity: Any) -> SortedSetData:
        if isinstance(entity, dict):
            return self.format_memory_data(entity['values'])

        return self.format_memory_data(entity.values)

    def format_memory_data(self, data: SortedSetData) -> SortedSetData:
        data = list(itertools.chain(*data))
        data.reverse()
        return data
